import enum

import cupy
import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OVERWRITE_ELEMENT,
    VECTOR_STRING_ELEMENT,
)

from . import base_calcat, base_gpu, utils
from ._version import version as deviceVersion
from .base_correction import BaseCorrection, add_correction_step_schema


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1


class DsscConstants(enum.Enum):
    Offset = enum.auto()


class DsscGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "dssc_gpu.cu"
    _corrected_axis_order = "cyx"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        self.input_shape = (memory_cells, pixels_y, pixels_x)
        self.processed_shape = self.input_shape
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )

        self.map_shape = (self.constant_memory_cells, self.pixels_y, self.pixels_x)
        self.offset_map_gpu = cupy.empty(self.map_shape, dtype=np.float32)

        self._init_kernels()

        self.offset_map_gpu = cupy.empty(self.map_shape, dtype=np.float32)

        self.update_block_size((1, 1, 64))

    def _get_raw_for_preview(self):
        return self.input_data_gpu

    def _get_corrected_for_preview(self):
        return self.processed_data_gpu

    def load_offset_map(self, offset_map):
        # can have an extra dimension for some reason
        if len(offset_map.shape) == 4:  # old format (see offsetcorrection_dssc.py)?
            offset_map = offset_map[..., 0]
        # shape (now): x, y, memory cell
        offset_map = np.transpose(offset_map).astype(np.float32)
        self.offset_map_gpu.set(offset_map)

    def correct(self, flags):
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                np.uint8(flags),
                self.offset_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")


class DsscCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = DsscConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            DsscConstants.Offset: self.dark_condition,
        }

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        super(DsscCalcatFriend, DsscCalcatFriend).add_schema(
            schema, managed_keys, "DSSC-Type", param_prefix, status_prefix
        )
        (
            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.memoryCells")
            .setNewDefaultValue(400)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.biasVoltage")
            .setNewDefaultValue(100)  # TODO: proper
            .commit()
        )

        base_calcat.add_status_schema_from_enum(schema, status_prefix, DsscConstants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        return res


@KARABO_CLASSINFO("DsscCorrection", deviceVersion)
class DsscCorrection(BaseCorrection):
    # subclass *must* set these attributes
    _correction_flag_class = CorrectionFlags
    _correction_field_names = (("offset", CorrectionFlags.OFFSET),)
    _kernel_runner_class = DsscGpuRunner
    _calcat_friend_class = DsscCalcatFriend
    _constant_enum_class = DsscConstants
    _managed_keys = BaseCorrection._managed_keys.copy()

    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.memoryCells")
            .setNewDefaultValue(400)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .setNewDefaultValue("cyx")
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("pulse")
            .commit(),
        )
        DsscCalcatFriend.add_schema(expected, DsscCorrection._managed_keys)
        add_correction_step_schema(
            expected,
            DsscCorrection._managed_keys,
            DsscCorrection._correction_field_names,
        )
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(DsscCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.get("dataFormat.memoryCells"),
            1,
            self.get("dataFormat.pixelsY"),
            self.get("dataFormat.pixelsX"),
        )

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        do_generate_preview,
    ):
        pulse_table = np.ravel(data_hash.get("image.pulseId"))
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            # drop the singleton dimension for kernel runner
            self.kernel_runner.load_data(image_data[:, 0])
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.load_cell_table(cell_table)
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        if do_generate_preview:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ) = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
                warn_func=self.log_status_warn,
            )
            preview_raw, preview_corrected = self.kernel_runner.compute_previews(
                preview_slice_index,
            )

        data_hash.set(self._image_data_path, buffer_handle)
        data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
        data_hash.set("image.pulseId", pulse_table[:, np.newaxis])
        data_hash.set("calngShmemPaths", [self._image_data_path])
        self._write_output(data_hash, metadata)
        if do_generate_preview:
            self._write_preview_outputs(
                (
                    ("preview.outputRaw", preview_raw),
                    ("preview.outputCorrected", preview_corrected),
                ),
                metadata,
            )

    def _load_constant_to_runner(self, constant, constant_data):
        assert constant is DsscConstants.Offset
        self.kernel_runner.load_offset_map(constant_data)
        if not self.get("corrections.offset.available"):
            self.set("corrections.offset.available", True)

        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to GPU")
